#include "runtime.h"
#include "lexer.h"
#include <stdlib.h>
#include <string.h>

void value_vec_construct(ValueVec* vec, Allocator* allocator)
{
    const size_t starting_capacity = 8;
    *vec = (ValueVec) {
        .allocator = allocator,
        .data = allocator->alloc(allocator, starting_capacity * sizeof(Value)),
        .length = 0,
        .capacity = starting_capacity,
    };
}

void value_vec_destroy(ValueVec* vec)
{
    for (size_t i = 0; i < vec->length; ++i) {
        value_destroy(&vec->data[i]);
    }
    free(vec->data);
}

Value* value_vec_at(const ValueVec* vec, int64_t index)
{
    if (index < 0) {
        return value_vec_at(vec, -index);
    }
    if (index >= vec->length) {
        return NULL;
    }
    return &vec->data[index];
}

void value_vec_reserve(ValueVec* vec, size_t min_capacity)
{
    if (vec->capacity < min_capacity + 1) {
        while (vec->capacity < min_capacity + 1) {
            vec->capacity *= 2;
        }
        vec->data = vec->allocator->realloc(
            vec->allocator, vec->data, vec->capacity * sizeof(Value)
        );
    }
}

void value_vec_push(ValueVec* vec, Value value)
{
    value_vec_reserve(vec, vec->length);
    vec->data[vec->length] = value;
    vec->length += 1;
}

Value value_vec_pop(ValueVec* vec)
{
    ASSERT(vec->length > 0);
    vec->length -= 1;
    return vec->data[vec->length];
}

ValueVec value_vec_clone(ValueVec* vec)
{
    ValueVec new_vec;
    value_vec_construct(&new_vec, vec->allocator);
    if (vec->length == 0) {
        return new_vec;
    }
    value_vec_reserve(&new_vec, vec->length - 1);
    new_vec.length = vec->length;
    memcpy(new_vec.data, vec->data, new_vec.length * sizeof(Value));
    return new_vec;
}

ValueVec value_vec_tail(ValueVec* vec)
{
    ValueVec new_vec;
    value_vec_construct(&new_vec, vec->allocator);
    if (vec->length <= 1) {
        return new_vec;
    }
    value_vec_reserve(&new_vec, vec->length - 1);
    new_vec.length = vec->length - 1;
    memcpy(new_vec.data, &vec->data[1], new_vec.length * sizeof(Value));
    return new_vec;
}

void value_vec_stringify(const ValueVec* vec, String* acc)
{
    string_append_cstr(acc, "[");
    if (vec->length > 0) {
        value_stringify(&vec->data[0], acc);
        for (size_t i = 1; i < vec->length; ++i) {
            string_append_cstr(acc, ", ");
            value_stringify(&vec->data[i], acc);
        }
    }
    string_append_cstr(acc, "]");
}

void value_destroy(Value* value)
{
    switch (value->type) {
        case ValueTypeSymbol:
            string_destroy(&value->symbol);
            break;
        case ValueTypeInt:
        case ValueTypeChar:
        case ValueTypeBool:
            break;
        case ValueTypeList:
            value_vec_destroy(&value->list);
            break;
    }
}

void value_stringify(const Value* value, String* acc)
{
    switch (value->type) {
        case ValueTypeSymbol:
            string_append(acc, &value->symbol);
            break;
        case ValueTypeInt:
            string_append_format(acc, "%ld", value->int_value);
            break;
        case ValueTypeChar:
            string_append_format(acc, "%c", value->char_value);
            break;
        case ValueTypeBool:
            string_append_cstr(acc, value->bool_value ? "true" : "false");
            break;
        case ValueTypeList:
            value_vec_stringify(&value->list, acc);
            break;
    }
}

void gc_buffer_construct(GCBuffer* buffer, size_t capacity)
{
    void* data = malloc(capacity);
    ASSERT(buffer != NULL);
    *buffer = (GCBuffer) {
        .data = data,
        .capacity = capacity,
    };
}

void gc_buffer_destroy(GCBuffer* buffer) { free(buffer->data); }

void gc_buffer_resize(GCBuffer* buffer, size_t minimum_capacity)
{
    if (buffer->capacity >= minimum_capacity) {
        return;
    }
    while (buffer->capacity < minimum_capacity) {
        buffer->capacity *= 2;
    }
    void* new_buffer = realloc(buffer->data, buffer->capacity);
    ASSERT(new_buffer);
    buffer->data = new_buffer;
}

void gc_construct(GC* allocator, size_t start_capacity)
{
    GCBuffer buffer_a;
    gc_buffer_construct(&buffer_a, start_capacity);
    GCBuffer buffer_b;
    gc_buffer_construct(&buffer_b, start_capacity);
    *allocator = (GC) {
        .alloc = gc_alloc,
        .realloc = gc_realloc,
        .buffer_a = buffer_a,
        .buffer_b = buffer_b,
        .select = GCBufferA,
        .index = 0,
    };
}

void gc_destroy(GC* allocator)
{
    gc_buffer_destroy(&allocator->buffer_a);
    gc_buffer_destroy(&allocator->buffer_b);
}

void* gc_alloc(GC* allocator, size_t size)
{
    GCBuffer* buffer = gc_selected_buffer(allocator);
    if (buffer->capacity < allocator->index + size) {
        return NULL;
    }
    void* ptr = &buffer->data[allocator->index];
    allocator->index += size;
    return ptr;
}

void* gc_realloc(GC* allocator, void* data, size_t size)
{
    void* new_ptr = gc_alloc(allocator, size);
    if (new_ptr == NULL) {
        return NULL;
    }
    memcpy(new_ptr, data, size);
    return new_ptr;
}

void* gc_collect(GC* allocator, ValueVec* stack, size_t to_be_allocated)
{
    GCBuffer* selected = gc_selected_buffer(allocator);
    GCBuffer* other = gc_other_buffer(allocator);
}

GCBuffer* gc_selected_buffer(GC* allocator)
{
    switch (allocator->select) {
        case GCBufferA:
            return &allocator->buffer_a;
        case GCBufferB:
            return &allocator->buffer_b;
    }
}

GCBuffer* gc_other_buffer(GC* allocator)
{
    switch (allocator->select) {
        case GCBufferA:
            return &allocator->buffer_b;
        case GCBufferB:
            return &allocator->buffer_a;
    }
}

void gc_select_other(GC* allocator)
{
    switch (allocator->select) {
        case GCBufferA:
            allocator->select = GCBufferB;
            break;
        case GCBufferB:
            allocator->select = GCBufferA;
            break;
    }
}

void runtime_construct(Runtime* runtime, Allocator* allocator)
{
    *runtime = (Runtime) {
        .allocator = allocator,
    };
}

void runtime_destroy(Runtime* runtime) { }
