#ifndef RUNTIME_H
#define RUNTIME_H

#include "lexer.h"
#include <stdbool.h>
#include <stdint.h>

typedef struct Allocator Allocator;
struct Allocator {
    void* (*alloc)(Allocator* allocator, size_t size);
    void* (*realloc)(Allocator* allocator, void* data, size_t size);
};

typedef enum {
    ValueTypeSymbol,
    ValueTypeInt,
    ValueTypeChar,
    ValueTypeBool,
    ValueTypeList,
} ValueType;

typedef struct Value Value;

typedef struct {
    Allocator* allocator;
    Value* data;
    size_t length;
    size_t capacity;
} ValueVec;

void value_vec_construct(ValueVec* vec, Allocator* allocator);
void value_vec_destroy(ValueVec* vec);
Value* value_vec_at(const ValueVec* vec, int64_t index);
void value_vec_reserve(ValueVec* vec, size_t min_capacity);
void value_vec_push(ValueVec* vec, Value value);
Value value_vec_pop(ValueVec* vec);
ValueVec value_vec_clone(ValueVec* vec);
ValueVec value_vec_tail(ValueVec* vec);
void value_vec_stringify(const ValueVec* vec, String* acc);

struct Value {
    ValueType type;
    union {
        String symbol;
        int64_t int_value;
        char char_value;
        bool bool_value;
        ValueVec list;
    };
};

void value_destroy(Value* value);
void value_stringify(const Value* value, String* acc);

typedef struct {
    void* data;
    size_t capacity;
} GCBuffer;

void gc_buffer_construct(GCBuffer* buffer, size_t capacity);
void gc_buffer_destroy(GCBuffer* buffer);
void gc_buffer_resize(GCBuffer* buffer, size_t minimum_capacity);

typedef enum {
    GCBufferA,
    GCBufferB,
} SelectedGCBuffer;

typedef struct GC GC;
struct GC {
    void* (*alloc)(GC* allocator, size_t size);
    void* (*realloc)(GC* allocator, void* data, size_t size);
    GCBuffer buffer_a;
    GCBuffer buffer_b;
    SelectedGCBuffer select;
    size_t index;
};

void gc_construct(GC* allocator, size_t start_capacity);
void gc_destroy(GC* allocator);
void* gc_alloc(GC* allocator, size_t size);
void* gc_realloc(GC* allocator, void* data, size_t size);
void* gc_collect(GC* allocator, ValueVec* stack, size_t to_be_allocated);

GCBuffer* gc_selected_buffer(GC* allocator);
GCBuffer* gc_other_buffer(GC* allocator);
void gc_select_other(GC* allocator);

typedef struct {
    Allocator* allocator;
} Runtime;

void runtime_construct(Runtime* runtime, Allocator* allocator);
void runtime_destroy(Runtime* runtime);

#endif
